@@ -1,36 +1,19 @@
 # pylint: disable=C0103
 from typing import Optional, Union
 import numpy as np
+from warnings import warn
 from pyrealm.utilities import summarize_attrs
 from pyrealm.param_classes import PModelParams
-from pyrealm.constrained_array import ConstraintFactory
+from pyrealm.bounds_checker import bounds_checker
 from pyrealm.utilities import check_input_shapes
 
 # TODO - Note that the typing currently does not enforce the dtype of ndarrays
 #        but it looks like the upcoming np.typing module might do this.
 
-# Define environmental variable constraint functions. These are intended
-# primarily to keep inputs away from areas where the functions become
-# numerically unstable.
-
-# TODO - check the sanity of these limits
-
-_constrain_temp = ConstraintFactory(label='temperature (°C)',
-                                    lower=0, upper=100)
-_constrain_patm = ConstraintFactory(label='atmospheric pressure (Pa)',
-                                    lower=30000, upper=110000)
-_constrain_vpd = ConstraintFactory(label='vapor pressure deficit (Pa)',
-                                   lower=0, upper=10000)
-_constrain_co2 = ConstraintFactory(label='carbon dioxide (ppm)',
-                                   lower=100, upper=1000)
-_constrain_elev = ConstraintFactory(label='elevation (m)',
-                                    lower=-500, upper=9000)
-
-
 def calc_density_h2o(tc: Union[float, np.ndarray],
                      patm: Union[float, np.ndarray],
-                     pmodel_params: PModelParams = PModelParams()
-                     ) -> Union[float, np.ndarray]:
+                     pmodel_params: PModelParams = PModelParams(),
+                     safe: bool = True) -> Union[float, np.ndarray]:
     """Calculates the **density of water** as a function of temperature and
     atmospheric pressure, using the Tumlirz Equation and coefficients calculated
     by :cite:`Fisher:1975tm`.
@@ -40,6 +23,8 @@ def calc_density_h2o(tc: Union[float, np.ndarray],
         tc: air temperature, °C
         patm: atmospheric pressure, Pa
         pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.
+        safe: Prevents the function from estimating density below -30°C, where the
+            function behaves poorly
 
     Other Parameters:
 
@@ -60,26 +45,14 @@ def calc_density_h2o(tc: Union[float, np.ndarray],
     # DESIGN NOTE:
     # It doesn't make sense to use this function for tc < 0, but in particular
     # the calculation shows wild numeric instability between -44 and -46 that
-    # leads to numerous downstream issues:
-    #
-    # from pyrealm import pmodel
-    # from matplotlib import pyplot as plt
-    # import numpy as np
-    #
-    # tc = np.arange(-88, 48, 0.1)
-    # r = np.arange(-1, 1, 0.01)
-    # sd = np.fromiter([np.std(pmodel.calc_density_h2o(t + r, 101350)) for t in tc], np.float)
-    #
-    # plt.plot(tc, sd)
-    # plt.show()
-
+    # leads to numerous downstream issues - see the extreme values documentation.
+    if safe and np.nanmin(tc) < -30:
+        raise RuntimeError('Water density calculations below about -30°C are '
+                           'unstable. See argument safe to calc_density_h2o')
+        
     # Check input shapes, shape not used
     _ = check_input_shapes(tc, patm)
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-    patm = _constrain_patm(patm)
-
     # Get powers of tc, including tc^0 = 1 for constant terms
     tc_pow = np.power.outer(tc, np.arange(0, 10))
 
@@ -226,9 +199,6 @@ def calc_ftemp_inst_rd(tc: Union[float, np.ndarray],
         250.9593
     """
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-
     return np.exp(pmodel_params.heskel_b * (tc - pmodel_params.k_To) -
                   pmodel_params.heskel_c * (tc ** 2 - pmodel_params.k_To ** 2))
 
@@ -289,9 +259,6 @@ def calc_ftemp_inst_vcmax(tc: Union[float, np.ndarray],
 
     """
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-
     # Convert temperatures to Kelvin
     tkref = pmodel_params.k_To + pmodel_params.k_CtoK
     tk = tc + pmodel_params.k_CtoK
@@ -362,9 +329,6 @@ def calc_ftemp_kphio(tc: Union[float, np.ndarray],
 
     """
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-
     if c4:
         coef = pmodel_params.kphio_C4
     else:
@@ -421,10 +385,6 @@ def calc_gammastar(tc: Union[float, np.ndarray],
     # check inputs, return shape not used
     _ = check_input_shapes(tc, patm)
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-    patm = _constrain_patm(patm)
-
     return (pmodel_params.bernacchi_gs25_0 * patm / pmodel_params.k_Po *
             calc_ftemp_arrh((tc + pmodel_params.k_CtoK), ha=pmodel_params.bernacchi_dha))
 
@@ -459,16 +419,12 @@ def calc_ns_star(tc: Union[float, np.ndarray],
 
     Examples:
 
-        >>> # Realative viscosity at 20 degrees Celsius and standard
+        >>> # Relative viscosity at 20 degrees Celsius and standard
         >>> # atmosphere (in Pa):
         >>> round(calc_ns_star(20, 101325), 5)
         1.12536
     """
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-    patm = _constrain_patm(patm)
-
     visc_env = calc_viscosity_h2o(tc, patm)
     visc_std = calc_viscosity_h2o(pmodel_params.k_To, pmodel_params.k_Po)
 
@@ -535,10 +491,6 @@ def calc_kmm(tc: Union[float, np.ndarray],
     # Check inputs, return shape not used
     _ = check_input_shapes(tc, patm)
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-    patm = _constrain_patm(patm)
-
     # conversion to Kelvin
     tk = tc + pmodel_params.k_CtoK
 
@@ -653,10 +605,6 @@ def calc_viscosity_h2o(tc: Union[float, np.ndarray],
     # Check inputs, return shape not used
     _ = check_input_shapes(tc, patm)
 
-    # Check input ranges
-    tc = _constrain_temp(tc)
-    patm = _constrain_patm(patm)
-
     # Get the density of water, kg/m^3
     rho = calc_density_h2o(tc, patm, pmodel_params=pmodel_params)
 
@@ -725,14 +673,6 @@ def calc_patm(elv: Union[float, np.ndarray],
     # in Kelvins, while other functions use this constant in the PARAM units of
     # °C.
 
-    # Check input ranges
-    # TODO - this leads to downstream issues. Using this in a calculation
-    #        results in constrained arrays without any of the attributes.
-    #        This is a wider implementation problem but currenly only impacts
-    #        here, where constrained arrays get passed into new calculations
-
-    # elv = _constrain_elev(elv)
-
     kto = pmodel_params.k_To + pmodel_params.k_CtoK
 
     return (pmodel_params.k_Po * (1.0 - pmodel_params.k_L * elv / kto) **
@@ -758,9 +698,6 @@ def calc_co2_to_ca(co2: Union[float, np.ndarray],
         41.850265
     """
 
-    # Check input ranges
-    co2 = _constrain_co2(co2)
-    patm = _constrain_patm(patm)
 
     return 1.0e-6 * co2 * patm  # Pa, atms. CO2
 
@@ -825,11 +762,19 @@ class PModelEnvironment:
 
         self.shape = check_input_shapes(tc, vpd, co2, patm)
 
-        # Store input variables
-        self.tc = _constrain_temp(tc)
-        self.vpd = _constrain_vpd(vpd)
-        self.co2 = _constrain_co2(co2)
-        self.patm = _constrain_patm(patm)
+        # Validate and store the forcing variables
+        self.tc = bounds_checker(tc, -25, 80, '[]', 'tc', '°C')
+        self.vpd = bounds_checker(vpd, 0, 10000, '[]', 'vpd', 'Pa')
+        self.co2 = bounds_checker(co2, 0, 1000, '[]', 'co2', 'ppm')
+        self.patm = bounds_checker(patm, 30000, 110000, '[]', 'tc', '°C')
+
+        # Guard against calc_density issues
+        if np.nanmin(self.tc) < -25:
+            raise ValueError('Cannot calculate P Model predictions for values below -25°C. See calc_density_h2o.')
+
+        # Guard against negative VPD issues
+        if np.nanmin(self.vpd) < 0:
+            raise ValueError('Negative VPD values will lead to missing data - clip to zero or explicitly set to np.nan')
 
         # ambient CO2 partial pressure (Pa)
         self.ca = calc_co2_to_ca(self.co2, self.patm)
@@ -874,32 +819,35 @@ class PModelEnvironment:
 class PModel:
 
     r"""Fits the P Model to a given set of environmental and photosynthetic
-    parameters. The calculated attributes of the class are
-    described below. An extended description with typical use cases is given
-    in :ref:`pmodel/pmodel.html` but the basic flow of the model is:
+    parameters. The calculated attributes of the class are described below. An
+    extended description with typical use cases is given in :any:`pmodel_overview`
+    but the basic flow of the model is:
 
     1. Estimate :math:`\ce{CO2}` limitation factors and optimal internal to
        ambient :math:`\ce{CO2}` partial pressure ratios (:math:`\chi`), using
        :class:`~pyrealm.pmodel.CalcOptimalChi`.
     2. Estimate photosynthetic efficiencies including water use efficiency,
-       light use efficiency (LUE) and maximum carboxylation rate (:math:`V_{cmax}`)
-       using :class:`~pyrealm.pmodel.CalcLUEVcmax`.
+       light use efficiency (LUE) and maximum carboxylation rate
+       (:math:`V_{cmax}`) using :class:`~pyrealm.pmodel.CalcLUEVcmax`.
     3. Optionally, estimate productivity measures including GPP by supplying
-       FAPAR and PPFD using the :meth:`~pyrealm.pmodel.PModel.estimate_productivity`
-       method.
+       FAPAR and PPFD using the
+       :meth:`~pyrealm.pmodel.PModel.estimate_productivity` method.
 
     **Corollary prediction details**
 
     These calculations use two additional functions:
 
-    * the instantaneous temperature response of :math:`V_{cmax}` (:math:`fv(t)`),
-      implemented in :func:`~pyrealm.pmodel.calc_ftemp_inst_vcmax`, and
+    * the instantaneous temperature response of :math:`V_{cmax}`
+      (:math:`fv(t)`), implemented in
+      :func:`~pyrealm.pmodel.calc_ftemp_inst_vcmax`, and
     * the instantaneous temperature response of dark respiration :math:`V_{d}`
-      (:math:`fr(t)`), implemented in :func:`~pyrealm.pmodel.calc_ftemp_inst_rd`.
+      (:math:`fr(t)`), implemented in
+      :func:`~pyrealm.pmodel.calc_ftemp_inst_rd`.
 
     The predictions are then:
 
-    * Intrinsic water use efficiency (iWUE, Pa), calculated as :math:`(c_a - c_i)/1.6`
+    * Intrinsic water use efficiency (iWUE, Pa), calculated as :math:`(c_a -
+      c_i)/1.6`
 
     * Maximum carboxylation capacity (mol C m-2) normalised to the standard
       temperature as: :math:`V_{cmax25} = V_{cmax}  / fv(t)`
@@ -910,7 +858,8 @@ class PModel:
 
             R_d = b_0 \frac{fr(t)}{fv(t)} V_{cmax}
 
-        following :cite:`Atkin:2015hk` (:math:`b_0` is set in `pmodel_params.atkin_rd_to_vcmax`)
+        following :cite:`Atkin:2015hk` (:math:`b_0` is set in
+        :attr:`~pyrealm.pmodel_params.atkin_rd_to_vcmax`)
 
     * Stomatal conductance (:math:`g_s`), calculated as:
 
@@ -918,7 +867,10 @@ class PModel:
 
             g_s = \frac{LUE}{M_C}\frac{1}{c_a - c_i}
 
-        When C4 photosynthesis is being used, :math:`g_s \to \infty`.
+        When C4 photosynthesis is being used, the true partial pressure of CO2
+        in the substomatal cavities (:math:`c_i`) is used following the
+        calculation of :math:`\chi` using
+        :attr:`~pyrealm.param_classes.PModelParams.beta_cost_ratio_c4`
 
     * The maximum rate of Rubsico regeneration at the growth temperature
       (:math:`J_{max}`) per unit irradiance is calculated as:
@@ -926,19 +878,20 @@ class PModel:
         .. math::
 
             J_{max} = \frac{4 \phi_0 I_{abs}}{\sqrt{\left(\frac{1}
-            {\left(\frac{V_{cmax}(c_i - 2 \Gamma^*)}
-            {\phi_0 I_{abs}(c_i + k_{mm})}\right)}\right)^2 - 1}}
+            {\left(\frac{V_{cmax}(c_i - 2 \Gamma^*)} {\phi_0 I_{abs}(c_i +
+            k_{mm})}\right)}\right)^2 - 1}}
 
     Parameters:
 
-        env: An instance of :class:`~pyrealm.pmodel.PModelEnvironment`.
-        kphio: (Optional) Apparent quantum yield efficiency (unitless).
+        env: An instance of :class:`~pyrealm.pmodel.PModelEnvironment`. 
+        kphio: (Optional) Apparent quantum yield efficiency (unitless). 
         rootzonestress: (Optional, default=None) An experimental option
             for providing a root zone water stress factor. This is not
             compatible with the soilmstress approach.
         soilmstress: (Optional, default=None) A soil moisture stress factor
             calculated using :func:`~pyrealm.pmodel.calc_soilmstress`.
-        c4: (Optional, default=False) Selects the C3 or C4 photosynthetic pathway.
+        c4: (Optional, default=False) Selects the C3 or C4 photosynthetic
+            pathway. 
         method_jmaxlim: (Optional, default=`wang17`) Method to use for
             :math:`J_{max}` limitation
         do_ftemp_kphio: (Optional, default=True) Include the temperature-
@@ -953,7 +906,7 @@ class PModel:
             (:class:`~pyrealm.param_classes.PModelParams`)
         optchi: Details of the optimal chi calculation
             (:class:`~pyrealm.pmodel.CalcOptimalChi`)
-        iwue: Intrinsic water use efficiency (iWUE, Pa)
+        iwue: Intrinsic water use efficiency (iWUE, Pa) 
         lue: Light use efficiency (LUE)
 
     Examples:
@@ -971,9 +924,9 @@ class PModel:
         >>> mod_c4 = PModel(env, c4=True, method_jmaxlim='none')
         >>> # Key variables from PModel
         >>> np.round(mod_c4.optchi.ci, 5)
-        40.53
+        18.22528
         >>> np.round(mod_c4.optchi.chi, 5)
-        1.0
+        0.44967
         >>> mod_c4.estimate_productivity(fapar=1, ppfd=300)
         >>> np.round(mod_c4.gpp, 5)
         103.25886
@@ -1175,9 +1128,10 @@ class PModel:
 
         # Check input shapes against each other and an existing calculated value
         _ = check_input_shapes(ppfd, fapar, self.lue)
-
+        
+        # Calculate Iabs 
         iabs = fapar * ppfd
-
+        
         # GPP
         self._gpp = self.lue * iabs
 
@@ -1195,26 +1149,34 @@ class PModel:
         self._rd = (self.pmodel_params.atkin_rd_to_vcmax *
                     (ftemp_inst_rd / ftemp25_inst_vcmax) * self.vcmax)
 
-        # Jmax using again A_J = A_C, handling edges cases
+        # Jmax using again A_J = A_C
         fact_jmaxlim = (self.vcmax * (self.optchi.ci + 2.0 * self.env.gammastar) /
-                        (self.kphio * (self.optchi.ci + self.env.kmm)))
-        fact_jmaxlim = (1.0 / fact_jmaxlim) ** 2 - 1.0
+                        (self.kphio * iabs * (self.optchi.ci + self.env.kmm)))
+        # Guard against negative values getting into sqrt
+        jmaxlim_step1 = (1.0 / fact_jmaxlim) ** 2 - 1.0
         jmax = np.empty_like(fact_jmaxlim)
-        mask = fact_jmaxlim > 0
-        jmax[mask] = 4.0 * self.kphio / np.sqrt(fact_jmaxlim[mask])
-        jmax[~ mask] = np.infty
+        mask = jmaxlim_step1 > 0
+        # Iabs might be a scalar or an array - if an array, it should be of the
+        # same shape as jmaxlim_step1 and can use the same mask
+        iabs_mask = iabs if isinstance(iabs, (float, int)) else iabs[mask]
+        jmax[mask] = 4.0 * self.kphio * iabs_mask / np.sqrt(jmaxlim_step1[mask])
+        jmax[~ mask] = np.nan
 
         # Revert to scalar if needed and store
         self._jmax = jmax.item() if np.ndim(jmax) == 0 else jmax
 
+        # AJ and AC
+        a_j = (self.kphio * iabs * (self.optchi.ci - self.env.gammastar) / 
+                (self.optchi.ci + 2 * self.env.gammastar) *  fact_jmaxlim)
+        a_c = (self.vcmax * (self.optchi.ci - self.env.gammastar)/ 
+                (self.optchi.ci + self.env.kmm))
+        assim = np.minimum(a_j, a_c)
+
+        if not np.allclose(assim, self._gpp / self.pmodel_params.k_c_molmass, equal_nan=True):
+            warn('Assimilation and GPP are not identical')
+
         # Stomatal conductance
-        if self.c4 and self.shape == 1:
-            self._gs = np.infty
-        elif self.c4:
-            self._gs = np.ones(self.shape) * np.infty
-        else:
-            self._gs = ((self.lue / self.pmodel_params.k_c_molmass) /
-                        (self.env.ca - self.optchi.ci))
+        self._gs = assim / (self.env.ca - self.optchi.ci)
 
     def __repr__(self):
         if self.do_soilmstress:
@@ -1306,14 +1268,16 @@ class CalcOptimalChi:
         0.7123
         >>> round(vals.mjoc, 5)
         2.13211
-        >>> # The c4 method just populates these values with 1.0
+        >>> # The c4 method estimates chi but sets the others at 1.
         >>> c4_vals = CalcOptimalChi(kmm = 46.09928, gammastar = 3.33925,
         ...                          ns_star = 1.12536, ca = 40.53, vpd = 1000,
         ...                          method='c4')
-        >>> c4_vals.chi
-        1.0
+        >>> round(c4_vals.chi, 5)
+        0.44967
     """
 
+    # TODO - move chi calc into __init__? Shared between the two methods
+
     def __init__(self,
                  kmm: Union[float, np.ndarray],
                  gammastar: Union[float, np.ndarray],
@@ -1357,24 +1321,37 @@ class CalcOptimalChi:
         return f"CalcOptimalChi(chi={self.chi}, ci={self.ci}, mc={self.mc}, mj={self.mj})"
 
     def c4(self, **kwargs):
-        r"""This method simply sets :math:`\chi = m_j = m_c = m_{joc} = 1.0` to
-        capture the boosted :math:`\ce{CO2}` concentrations at the chloropolast in C4
+        r"""Optimal :math:`\chi` is calculated following Equation 8 in
+        :cite:`Prentice:2014bc` (see :meth:`~pyrealm.pmodel.CalcOptimalChi.prentice14`),
+        but using a C4 specific estimate of the unit cost ratio :math:`\beta`,
+        specified in :meth:`~pyrealm.param_classes.PModelParams.beta_cost_ratio_c4`.
+
+        This method then simply sets :math:`m_j = m_c = m_{joc} = 1.0` to capture the
+        boosted :math:`\ce{CO2}` concentrations at the chloropolast in C4
         photosynthesis.
         """
 
-        # Dummy values to represent c4 pathway. These need to retain any
+        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
+        xi = np.sqrt((self.pmodel_params.beta_cost_ratio_c4 *
+                      kwargs['rootzonestress'] *
+                      (kwargs['kmm'] + kwargs['gammastar']))
+                     / (1.6 * kwargs['ns_star']))
+
+        self.chi = (kwargs['gammastar'] / kwargs['ca'] +
+                    (1.0 - kwargs['gammastar'] / kwargs['ca']) * xi /
+                    (xi + np.sqrt(kwargs['vpd'])))
+
+        # These values need to retain any
         # dimensions of the original inputs - if ftemp_kphio is set to 1.0
         # (i.e. no temperature correction) then the dimensions of tc are lost
         # and the input to soilmstress might be scalar, so enforce the shape.
         # Note that rpmodel_1.0.6 collapses array inputs at this point.
 
         if self.shape == 1:
-            self.chi = 1.0
             self.mc = 1.0
             self.mj = 1.0
             self.mjoc = 1.0
         else:
-            self.chi = np.ones(self.shape)
             self.mc = np.ones(self.shape)
             self.mj = np.ones(self.shape)
             self.mjoc = np.ones(self.shape)
@@ -1415,21 +1392,20 @@ class CalcOptimalChi:
         assimilation.
         """
 
-        # Avoid negative VPD (dew conditions)
-        vpd = np.clip(kwargs['vpd'], 0, None)
-
         # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
-        xi = np.sqrt((self.pmodel_params.stocker19_beta *
+        xi = np.sqrt((self.pmodel_params.beta_cost_ratio_c3 *
                       kwargs['rootzonestress'] *
                       (kwargs['kmm'] + kwargs['gammastar']))
                      / (1.6 * kwargs['ns_star']))
+
         self.chi = (kwargs['gammastar'] / kwargs['ca'] +
-                    (1.0 - kwargs['gammastar'] / kwargs['ca']) * xi / (xi + np.sqrt(vpd)))
+                    (1.0 - kwargs['gammastar'] / kwargs['ca']) * xi /
+                    (xi + np.sqrt(kwargs['vpd'])))
 
         # Define variable substitutes:
         vdcg = kwargs['ca'] - kwargs['gammastar']
         vacg = kwargs['ca'] + 2.0 * kwargs['gammastar']
-        vbkg = (self.pmodel_params.stocker19_beta *
+        vbkg = (self.pmodel_params.beta_cost_ratio_c3 *
                 kwargs['rootzonestress'] *
                 (kwargs['kmm'] + kwargs['gammastar']))
 
@@ -1438,7 +1414,7 @@ class CalcOptimalChi:
         # unwrapped matrix inputs. Also, rpmodel includes a check for vpd > 0,
         # but this is guaranteed by clip above (also true in rpmodel).
 
-        vsr = np.sqrt(1.6 * kwargs['ns_star'] * vpd / vbkg)
+        vsr = np.sqrt(1.6 * kwargs['ns_star'] * kwargs['vpd'] / vbkg)
         mj = vdcg / (vacg + 3.0 * kwargs['gammastar'] * vsr)
 
         # Mask values with ns star <= 0 and vbkg <=0
@@ -1455,7 +1431,6 @@ class CalcOptimalChi:
         self.mc = (self.chi - gamma) / (self.chi + kappa)
         self.mjoc = (self.chi + kappa) / (self.chi + 2 * gamma)
 
-
     def summarize(self, dp=2):
         """Prints a summary of the variables calculated within an instance
         of CalcOptimalChi including the mean, range and number of nan values.
